@@ -1,4 +1,5 @@
 import math
+import types
 from datetime import datetime, timedelta
 from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union
 
@@ -19,8 +20,10 @@ from sqlalchemy import (
 from sqlalchemy.ext.declarative import declared_attr, has_inherited_table
 from sqlalchemy.ext.hybrid import hybrid_method, hybrid_property
 from sqlalchemy.orm import Session, backref, relationship
+from sqlalchemy.orm.util import AliasedClass
 from sqlalchemy.schema import UniqueConstraint
 from sqlalchemy.sql.elements import BinaryExpression
+from sqlalchemy.sql.expression import Selectable
 
 import timely_beliefs.utils as tb_utils
 from timely_beliefs.beliefs import probabilistic_utils
@@ -36,6 +39,12 @@ from timely_beliefs.visualization import utils as visualization_utils
 METADATA = ["sensor", "event_resolution"]
 DatetimeLike = Union[datetime, str, pd.Timestamp]
 TimedeltaLike = Union[timedelta, str, pd.Timedelta]
+JoinTarget = Union[
+    Selectable,
+    type,
+    AliasedClass,
+    types.FunctionType,
+]
 
 
 class TimedBelief(object):
@@ -236,7 +245,7 @@ class TimedBeliefDBMixin(TimedBelief):
     def add_to_session(
         cls,
         session: Session,
-        beliefs_data_frame: "BeliefsDataFrame",
+        beliefs_data_frame: "BeliefsDataFrame",  # todo: allow passing List[TimedBelief], too
         expunge_session: bool = False,
         allow_overwrite: bool = False,
         bulk_save_objects: bool = False,
@@ -282,6 +291,22 @@ class TimedBeliefDBMixin(TimedBelief):
         if commit_transaction:
             session.commit()
 
+    @classmethod
+    def expunge_from_session(
+        cls,
+        session: Session,
+        beliefs_data_frame: "BeliefsDataFrame",  # todo: allow passing List[TimedBelief], too
+    ):
+        # Belief timing is stored as the belief horizon rather than as the belief time
+        belief_records = (
+            beliefs_data_frame.convert_index_from_belief_time_to_horizon()
+                .reset_index()
+                .to_dict("records")
+        )
+        beliefs = [cls(sensor=beliefs_data_frame.sensor, **d) for d in belief_records]
+        for belief in beliefs:
+            session.expunge(belief)
+
     @classmethod
     @tb_utils.append_doc_of("TimedBeliefDBMixin.search_session")
     def query(cls, *args, **kwargs):
@@ -314,14 +339,17 @@ class TimedBeliefDBMixin(TimedBelief):
         source: Optional[Union[BeliefSource, List[BeliefSource]]] = None,
         most_recent_beliefs_only: bool = False,
         most_recent_events_only: bool = False,
-        most_recent_only: bool = False,  # deprecated
+        most_recent_only: bool = None,  # deprecated
         place_beliefs_in_sensor_timezone: bool = True,
         place_events_in_sensor_timezone: bool = True,
-        custom_filter_criteria: List[BinaryExpression] = None,
+        custom_filter_criteria: Optional[List[BinaryExpression]] = None,
+        custom_join_targets: Optional[List[JoinTarget]] = None,
     ) -> "BeliefsDataFrame":
         """Search a database session for beliefs about sensor events.
 
-        The optional arguments represent optional filters.
+        The optional arguments represent optional filters, with two exceptions:
+        - sensor_class makes it possible to create a query on sensor subclasses
+        - custom_join_targets makes it possible to add custom filters using other (incl. subclassed) targets
         :param session: the database session to use
         :param sensor: sensor to which the beliefs pertain, or its unique sensor id
         :param sensor_class: optionally pass the sensor (sub)class explicitly (only needed if you pass a sensor id instead of a sensor, and your sensor class is not DBSensor); the class should be mapped to a database table
@@ -336,7 +364,8 @@ class TimedBeliefDBMixin(TimedBelief):
         :param most_recent_events_only: only return (post knowledge time) beliefs for the most recent event (maximum event start)
         :param place_beliefs_in_sensor_timezone: if True (the default), belief times are converted to the timezone of the sensor
         :param place_events_in_sensor_timezone: if True (the default), event starts are converted to the timezone of the sensor
-        :param custom_filter_criteria: optionally pass additional filters, such as ones that rely on subclasses
+        :param custom_filter_criteria: additional filters, such as ones that rely on subclasses
+        :param custom_join_targets: additional join targets, to accommodate filters that rely on other targets (e.g. subclasses)
         :returns: a multi-index DataFrame with all relevant beliefs
         """
 
@@ -501,9 +530,12 @@ class TimedBeliefDBMixin(TimedBelief):
                 ),
             )
 
-        # Apply custom filter criteria
+        # Apply custom filter criteria and join targets
         if custom_filter_criteria is not None:
             q = q.filter(*custom_filter_criteria)
+        if custom_join_targets is not None:
+            for target in custom_join_targets:
+                q = q.join(target)
 
         # Build our DataFrame of beliefs
         df = BeliefsDataFrame(sensor=sensor, beliefs=q.all())
@@ -1031,8 +1063,8 @@ class BeliefsDataFrame(pd.DataFrame):
         self, fnc: Callable = None, *args: Any, **kwargs: Any
     ) -> Union["BeliefsDataFrame", DataFrameGroupBy]:
         """Convenient function to apply a function to each belief in the BeliefsDataFrame.
-        A belief is a group with unique event start, belief time and source. A deterministic belief is defined by a
-        single row, whereas a probabilistic belief is defined by multiple rows.
+        A belief is a group with unique event start, belief time and source.
+        Each individual belief may be deterministic (defined by a single row), or probabilistic (multiple rows).
         If no function is given, return the GroupBy object.
 
         :Example:
@@ -1044,6 +1076,15 @@ class BeliefsDataFrame(pd.DataFrame):
         >>> # If you want to call this method within another groupby function, pass the df group explicitly
         >>> df.for_each_belief(some_function, True, a=1, df=df)
         """
+        return self._for_each_belief(fnc, False, *args, **kwargs)
+
+    def _for_each_belief(
+        self, fnc: Callable, collective_beliefs: bool, *args: Any, **kwargs: Any
+    ) -> Union["BeliefsDataFrame", DataFrameGroupBy]:
+        """
+        If collective_beliefs is True, just group by event start and belief time.
+        Otherwise, group beliefs by source, too.
+        """
         df = kwargs.pop("df", self)
         index_names = []
         index_names.extend(
@@ -1060,12 +1101,33 @@ class BeliefsDataFrame(pd.DataFrame):
             if "belief_horizon" in df.index.names
             else []
         )
-        index_names.append("source")
+        if collective_beliefs is False:
+            index_names.append("source")
         gr = df.groupby(level=index_names, group_keys=False)
         if fnc is not None:
             return gr.apply(lambda x: fnc(x, *args, **kwargs))
         return gr
 
+    @hybrid_method
+    def for_each_collective_belief(
+        self, fnc: Callable = None, *args: Any, **kwargs: Any
+    ) -> Union["BeliefsDataFrame", DataFrameGroupBy]:
+        """Convenient function to apply a function to each collective belief in the BeliefsDataFrame.
+        A collective belief is a group with unique event start and belief time, which may contain multiple sources.
+        Each individual belief may be deterministic (defined by a single row), or probabilistic (multiple rows).
+        If no function is given, return the GroupBy object.
+
+        :Example:
+
+        >>> # Apply some function that accepts a DataFrame, a positional argument and a keyword argument
+        >>> df.for_each_collective_belief(some_function, True, a=1)
+        >>> # Pipe some other function that accepts a GroupBy object, a positional argument and a keyword argument
+        >>> df.for_each_collective_belief().pipe(some_other_function, True, a=1)
+        >>> # If you want to call this method within another groupby function, pass the df group explicitly
+        >>> df.for_each_collective_belief(some_function, True, a=1, df=df)
+        """
+        return self._for_each_belief(fnc, True, *args, **kwargs)
+
     @hybrid_method
     def belief_history(
         self,
@@ -1284,11 +1346,16 @@ class BeliefsDataFrame(pd.DataFrame):
         )
 
         # fast track a common case where each event has only one deterministic belief and only the most recent belief is needed
+        print("bbbbbbbbb")
+        print(df.lineage.number_of_beliefs == df.lineage.number_of_events)
+        print(keep_only_most_recent_belief)
+        print( df.lineage.number_of_sources)
         if (
             df.lineage.number_of_beliefs == df.lineage.number_of_events
             and keep_only_most_recent_belief
             and df.lineage.number_of_sources == 1
         ):
+            print("rrrrrrrrrr")
             if event_resolution > self.event_resolution:
                 # downsample
                 column_functions = {
@@ -1362,12 +1429,12 @@ class BeliefsDataFrame(pd.DataFrame):
         """Simply get the accuracy of beliefs about events, at a given time (pass a datetime), at a given horizon
         (pass a timedelta), or as a function of horizon (the default).
 
-        By default the accuracy is determined with respect to the most recent beliefs held by the same source.
+        By default, the accuracy is determined with respect to the most recent beliefs held by the same source.
         Optionally, set a reference source to determine accuracy with respect to beliefs held by a specific source.
 
-        By default the accuracy is determined with respect to the most recent beliefs.
+        By default, the accuracy is determined with respect to the most recent beliefs.
 
-        By default the mean absolute error (MAE), the mean absolute percentage error (MAPE) and
+        By default, the mean absolute error (MAE), the mean absolute percentage error (MAPE) and
         the weighted absolute percentage error (WAPE) are returned.
 
         For more options, use df.fixed_viewpoint_accuracy() or df.rolling_viewpoint_accuracy() instead.
@@ -1429,7 +1496,7 @@ class BeliefsDataFrame(pd.DataFrame):
         Alternatively, select the accuracy of beliefs formed within a certain time window. This allows setting a maximum
         acceptable freshness of the data.
 
-        By default the accuracy is determined with respect to the reference values in the `reference_value` column.
+        By default, the accuracy is determined with respect to the reference values in the `reference_value` column.
         This column is created if it does not exist, or if one of the following reference parameters is not None.
         By default, the reference values are the most recent beliefs held by the same source.
         - Optionally, set a reference belief time to determine accuracy with respect to beliefs at a specific time.
@@ -1437,7 +1504,7 @@ class BeliefsDataFrame(pd.DataFrame):
         - Optionally, set a reference source to determine accuracy with respect to beliefs held by a specific source.
         These options allow to define what is considered to be true at a certain time.
 
-        By default the mean absolute error (MAE), the mean absolute percentage error (MAPE) and
+        By default, the mean absolute error (MAE), the mean absolute percentage error (MAPE) and
         the weighted absolute percentage error (WAPE) are returned.
 
         :Example:
@@ -1512,14 +1579,14 @@ class BeliefsDataFrame(pd.DataFrame):
         knowledge time (with negative horizons indicating post knowledge time).
         This allows setting a maximum acceptable freshness of the data.
 
-        By default the accuracy is determined with respect to the reference values in the `reference_value` column.
+        By default, the accuracy is determined with respect to the reference values in the `reference_value` column.
         This column is created if it does not exist, or if one of the following reference parameters is not None.
         By default, the reference values are the most recent beliefs held by the same source.
         - Optionally, set a reference belief horizon to determine accuracy with respect to beliefs at a specific horizon.
         - Optionally, set a reference source to determine accuracy with respect to beliefs held by a specific source.
-        These options allow to define what is considered to be true at a certain time after an event.
+        These options allow defining what is considered to be true at a certain time after an event.
 
-        By default the mean absolute error (MAE), the mean absolute percentage error (MAPE) and
+        By default, the mean absolute error (MAE), the mean absolute percentage error (MAPE) and
         the weighted absolute percentage error (WAPE) are returned.
 
         :Example:
@@ -1677,14 +1744,15 @@ class BeliefsDataFrame(pd.DataFrame):
         reference_belief_time: datetime = None,
         reference_belief_horizon: timedelta = None,
         reference_source: BeliefSource = None,
-        return_expected_value: bool = False,
+        return_reference_type: str = "full",
+        return_expected_value: Optional[bool] = None,  # deprecated
     ) -> "BeliefsDataFrame":
         """Add a column with reference values.
-        By default the reference will be the expected value of the most recent belief held by the same source.
+        By default, the reference will be the probabilistic value of the most recent belief held by the same source.
+        To set a deterministic reference, use either return_expected_value or return_middle_value.
         Optionally, set a reference belief horizon.
         Optionally, set a reference source present in the BeliefsDataFrame.
-        These options allow to define what is considered to be true at a certain time after an event.
-        Todo: Option to set probabilistic reference values
+        These options allow defining what is considered to be true at a certain time after an event.
 
         :param reference_belief_time: optional datetime to indicate that
                the accuracy should be determined with respect to the latest belief held at this time
@@ -1692,9 +1760,23 @@ class BeliefsDataFrame(pd.DataFrame):
                the accuracy should be determined with respect to the latest belief at this duration past knowledge time
         :param reference_source: optional BeliefSource to indicate that
                the accuracy should be determined with respect to the beliefs held by the given source
-        :param return_expected_value: if True, set a deterministic reference by picking the expected value
+        :param return_reference_type: valid strings are:
+               - "full": a probabilistic reference using the full distribution
+               - "mean": a deterministic reference using the mean value
+               - "median": a deterministic reference using the median value
         """
 
+        # todo: deprecate the 'return_expected_value' argument in favor of 'return_reference_type' (announced v1.9.0)
+        return_expected_value = tb_utils.replace_deprecated_argument(
+            "return_expected_value",
+            return_expected_value,
+            "return_reference_type",
+            return_reference_type,
+            required_argument=False,
+        )
+        if isinstance(return_expected_value, bool):
+            return_reference_type = "mean" if return_expected_value else "full"
+
         df = self
 
         reference_df = df.groupby(level="event_start").apply(
@@ -1703,12 +1785,12 @@ class BeliefsDataFrame(pd.DataFrame):
                 reference_belief_time=reference_belief_time,
                 reference_belief_horizon=reference_belief_horizon,
                 reference_source=reference_source,
-                return_expected_value=return_expected_value,
+                return_reference_type=return_reference_type,
             )
         )
 
         # Concat to add a column while keeping original cp index level
-        if return_expected_value is True:
+        if return_reference_type != "full":
             if "belief_time" in df.index.names:
                 df = df.convert_index_from_belief_time_to_horizon()
                 df = pd.concat(
